/**************************************************************************
 * VGA mode 13h Graphics   Victor L. Miller                               *
 *                                                                        *
 * Use of VGA mode 13h Graphics                                           *
 * Remember to compile in the LARGE memory model!                         *
 *                                                                        *
 * This program will only work on DOS- or Windows-based systems with a    *
 * VGA, SuperVGA or compatible video adapter.                             *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <math.h>
#include <alloc.h>
#include <mem.h>

#define VIDEO_INT           0x10      /* the BIOS video interrupt. */
#define WRITE_DOT           0x0C      /* BIOS func to plot a pixel. */
#define SET_MODE            0x00      /* BIOS func to set the video mode. */
#define VGA_256_COLOR_MODE  0x13      /* use to set 256-color mode. */
#define TEXT_MODE           0x03      /* use to set 80x25 text mode. */

#define SCREEN_WIDTH        320       /* width in pixels  of mode 0x13 */
#define SCREEN_HEIGHT       200       /* height in pixels of mode 0x13 */
#define NUM_COLORS          256       /* number of colors in mode 0x13 */

#define sgn(x) ((x<0)?-1:((x>0)?1:0)) /* macro to return the sign of a
                                         number */
typedef long           fixed16_16;

fixed16_16 SIN_ACOS[1024];
float pi=3.141592;                     /* PI = 3.141592 ................ */

typedef unsigned char  byte;
typedef unsigned short word;

char far *VGA=(char far *)MK_FP(0xA000,0x0000);      /* This Points To Video Memory */
char far *my_clock=(char far *)MK_FP(0x0000,0x046C); /* This Points To The 18.2hz System Clock */

float x,y,z,xa,ya,za;
float r1,r2,r3,rd1,rd2,rd3;
float sr1, sr2, sr3;
float cr1, cr2, cr3;
float sx,sy,sz,sxa,sya,sxb,syb;
float d=700.,movex=60.,movey=60.,movez=1200.;

/**************************************************************************
 *  set_mode                                                              *
 *  Sets the video mode.                                                  *
 **************************************************************************/

void set_mode(byte mode)
{
  union REGS regs;

  regs.h.ah = SET_MODE;
  regs.h.al = mode;
  int86(VIDEO_INT, &regs, &regs);
}

/**************************************************************************
 *  plot_pixel_slow                                                       *
 *    Plot a pixel by using BIOS function 0x0C (Write Dot).               *
 **************************************************************************/

void plot_pixel_slow(int x,int y,byte color)
{
  union REGS regs;

  regs.h.ah = WRITE_DOT;
  regs.h.al = color;
  regs.x.cx = x;
  regs.x.dx = y;
  int86(VIDEO_INT, &regs, &regs);
}

/**************************************************************************
 *  plot_pixel_fast                                                       *
 *    Plot a pixel by directly writing to video memory.                   *
 **************************************************************************/

void plot_pixel_fast(int x,int y,byte color)
{
   VGA[y*SCREEN_WIDTH+x]=color;
}

/**************************************************************************
 *  plot2pixel2fast                                                       *
 *    Plot a pixel by directly writing to video memory.                   *
 **************************************************************************/

void plot2pixel2fast(int x,int y,byte color, char far *screen)
{
   screen[y*SCREEN_WIDTH+x]=color;
}

/**************************************************************************
 *  line_slow                                                             *
 *    draws a line using multiplication and division.                     *
 **************************************************************************/

void line_slow(int x1, int y1, int x2, int y2, byte color)
{
  int dx,dy,sdx,sdy,px,py,dxabs,dyabs,i;
  float slope;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  if (dxabs>=dyabs) /* the line is more horizontal than vertical */
  {
    slope=(float)dy / (float)dx;
    for(i=0;i!=dx;i+=sdx)
    {
      px=i+x1;
      py=slope*i+y1;
      plot_pixel_fast(px,py,color);
    }
  }
  else /* the line is more vertical than horizontal */
  {
    slope=(float)dx / (float)dy;
    for(i=0;i!=dy;i+=sdy)
    {
      px=slope*i+x1;
      py=i+y1;
      plot_pixel_fast(px,py,color);
    }
  }
}

/**************************************************************************
 *  line_fast                                                             *
 *    draws a line using Bresenham's line-drawing algorithm, which uses   *
 *    no multiplication or division.                                      *
 **************************************************************************/

void line_fast(int x1, int y1, int x2, int y2, byte color)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;

   VGA[(py<<8)+(py<<6)+px]=color;

  if (dxabs>=dyabs) /* the line is more horizontal than vertical */
  {
    for(i=0;i<dxabs;i++)
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
      }
      px+=sdx;
      plot_pixel_fast(px,py,color);
    }
  }
  else /* the line is more vertical than horizontal */
  {
    for(i=0;i<dyabs;i++)
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
      }
      py+=sdy;
      plot_pixel_fast(px,py,color);
    }
  }
}

/**************************************************************************
 *  line2fast                                                             *
 *    draws a line using Bresenham's line-drawing algorithm, which uses   *
 *    no multiplication or division.                                      *
 **************************************************************************/

void line2fast(int x1, int y1, int x2, int y2, byte color,
               char far *screen)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;

   screen[(py<<8)+(py<<6)+px]=color;

  if (dxabs>=dyabs) /* the line is more horizontal than vertical */
  {
    for(i=0;i<dxabs;i++)
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
      }
      px+=sdx;
      plot2pixel2fast(px,py,color,screen);
    }
  }
  else /* the line is more vertical than horizontal */
  {
    for(i=0;i<dyabs;i++)
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
      }
      py+=sdy;
      plot2pixel2fast(px,py,color,screen);
    }
  }
}

/**************************************************************************
 *  circle_slow                                                           *
 *    Draws a circle by using floating point numbers and math fuctions.   *
 **************************************************************************/

void circle_slow(int x,int y, int radius, byte color)
{
  float n=0,invradius=1/(float)radius;
  int dx=0,dy=radius-1;
  word dxoffset,dyoffset,offset=(y<<8)+(y<<6)+x;

  while (dx<=dy)
  {
    dxoffset = (dx<<8) + (dx<<6);
    dyoffset = (dy<<8) + (dy<<6);
    VGA[offset+dy-dxoffset] = color;  /* octant 0 */
    VGA[offset+dx-dyoffset] = color;  /* octant 1 */
    VGA[offset-dx-dyoffset] = color;  /* octant 2 */
    VGA[offset-dy-dxoffset] = color;  /* octant 3 */
    VGA[offset-dy+dxoffset] = color;  /* octant 4 */
    VGA[offset-dx+dyoffset] = color;  /* octant 5 */
    VGA[offset+dx+dyoffset] = color;  /* octant 6 */
    VGA[offset+dy+dxoffset] = color;  /* octant 7 */
    dx++;
    n+=invradius;
    dy=radius * sin(acos(n));
  }
}

/**************************************************************************
 *  circle_fast                                                           *
 *    Draws a circle by using fixed point numbers and a trigonometry      *
 *    table.                                                              *
 **************************************************************************/

void circle_fast(int x,int y, int radius, byte color)
{
  fixed16_16 n=0,invradius=(1/(float)radius)*0x10000L;
  int dx=0,dy=radius-1;
  word dxoffset,dyoffset,offset = (y<<8)+(y<<6)+x;

  while (dx<=dy)
  {
    dxoffset = (dx<<8) + (dx<<6);
    dyoffset = (dy<<8) + (dy<<6);
    VGA[offset+dy-dxoffset] = color;  /* octant 0 */
    VGA[offset+dx-dyoffset] = color;  /* octant 1 */
    VGA[offset-dx-dyoffset] = color;  /* octant 2 */
    VGA[offset-dy-dxoffset] = color;  /* octant 3 */
    VGA[offset-dy+dxoffset] = color;  /* octant 4 */
    VGA[offset-dx+dyoffset] = color;  /* octant 5 */
    VGA[offset+dx+dyoffset] = color;  /* octant 6 */
    VGA[offset+dy+dxoffset] = color;  /* octant 7 */
    dx++;
    n+=invradius;
    dy = (radius * SIN_ACOS[n>>6]) >>16;
  }
}

/**************************************************************************
 *  circle_fill                                                           *
 *    Draws and fills a circle.                                           *
 **************************************************************************/

void circle_fill(int x,int y, int radius, byte color)
{
  fixed16_16 n=0,invradius=(1/(float)radius)*0x10000L;
  int dx=0,dy=radius-1,i;
  word dxoffset,dyoffset,offset = (y<<8)+(y<<6)+x;

  while (dx<=dy)
  {
    dxoffset = (dx<<8) + (dx<<6);
    dyoffset = (dy<<8) + (dy<<6);
    for(i=dy;i>=dx;i--,dyoffset-=SCREEN_WIDTH)
    {
      VGA[offset+i -dxoffset] = color;  /* octant 0 */
      VGA[offset+dx-dyoffset] = color;  /* octant 1 */
      VGA[offset-dx-dyoffset] = color;  /* octant 2 */
      VGA[offset-i -dxoffset] = color;  /* octant 3 */
      VGA[offset-i +dxoffset] = color;  /* octant 4 */
      VGA[offset-dx+dyoffset] = color;  /* octant 5 */
      VGA[offset+dx+dyoffset] = color;  /* octant 6 */
      VGA[offset+i +dxoffset] = color;  /* octant 7 */
    }
    dx++;
    n+=invradius;
    dy = (radius * SIN_ACOS[n>>6]) >>16;
  }
}

/************************************************************************
 * SUBROUTINE KEYBOARD()    -  CHECK THE KEYBOARD BUFFER                 *
 *                             HOLD UNTIL <Esc> IS PRESSED               *
 *************************************************************************/
 void keyboard(void) {
 union u_type {int a; char b[3];} keystroke;
 int get_keystroke(void);

 do keystroke.a = get_keystroke();        /* Go To Local Subroutine */
 while (keystroke.b[0]!=27);       }      /* 27 Ascii is <Esc> key  */

 /* Local Subroutine: Retrieve One Keystroke       */
 int get_keystroke(void) {
 union REGS regs;
 regs.h.ah=0;
 return int86(0x16, &regs, &regs); }

 /************************************************************************
 * SUBROUTINE KEYBOARD_INPUT()    -  CHECK THE KEYBOARD BUFFER           *
 *                                   HOLD UNTIL <Esc> IS PRESSED         *
 *************************************************************************/

void keyboard_input(void){
union u_type{int a; char b[3];}keystroke; char inkey=0;

if(bioskey(1)==0) return;                   /* if no key, return      */
keystroke.a=bioskey(0);                     /* fetch ASCII code...    */
inkey=keystroke.b[0];                       /* ... and load code into variable */
switch (inkey) {                            /* make decision based upon ASCII value */
case 27:  exit(0);                          /* exit on <Esc> key      */
case 61:  movez=movez+20; return;           /* + key = move outwards  */
case 45:  movez=movez-20; return;           /* - key = move inwards   */
case 104: rd1=rd1-2;      return;           /* h key roll right       */
case 102: rd1=rd1+2;      return;           /* f key roll left        */
case 103: rd1=0;rd2=0;rd3=0; return;        /* g key return to nominal position */
case 116: rd2=rd2+2;      return;           /* t key pitch forward    */
case 98:  rd2=rd2-2;      return;           /* b key pitch back       */

default: return;} }                         /* make routine bullet-proof */

/***************************************************************
*       SUBROUTINE: CALCULATE SIN, COS FACTORS                 *
*  Enter with r1,r2,r3 viewing angles for yaw, roll, pitch     *
*  expressed in radians.  Returns sine and cosine factors.     *
****************************************************************/
void rotation(void){
sr1=sin(r1); sr2=sin(r2); sr3=sin(r3);
cr1=cos(r1); cr2=cos(r2); cr3=cos(r3); return;}


/******************************************************************
*       SUBROUTINE: STANDARD 3D FORMULAS                          *
*   Enter with x,y,z cartesian world coordinates.  Returns sx,sy  *
*   cartesian display coordinates.  Returns x,y,z cartesian view  *
*   coordinates which includes rotation and translation effects.  *
*******************************************************************/

void calc_3d(void){
x=(-1)*x; xa=cr1*x-sr1*z; za=sr1*x+cr1*z; x=cr2*xa+sr2*y;
ya=cr2*y-sr2*xa; z=cr3*za-sr3*ya; y=sr3*za+cr3*ya;
x=x+movex; y=y+movey; z=z+movez;
sx=d*x/z; sy=d*y/z; return;}
